#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <unistd.h>
#include <RF24/RF24.h>
#include <time.h>

using namespace std;
/****************** Raspberry Pi ***********************/
// Radio CE Pin, CSN Pin, SPI Speed
// See http://www.airspayce.com/mikem/bcm2835/group__constants.html#ga63c029bd6500167152db4e57736d0939 and the related enumerations for pin information.

// Setup for GPIO 22 CE and CE0 CSN with SPI Speed @ 4Mhz
//RF24 radio(RPI_V2_GPIO_P1_22, BCM2835_SPI_CS0, BCM2835_SPI_SPEED_4MHZ);

// NEW: Setup for RPi B+
//RF24 radio(RPI_BPLUS_GPIO_J8_15,RPI_BPLUS_GPIO_J8_24, BCM2835_SPI_SPEED_8MHZ);

// Setup for GPIO 15 CE and CE0 CSN with SPI Speed @ 8Mhz
//RF24 radio(RPI_V2_GPIO_P1_15, RPI_V2_GPIO_P1_24, BCM2835_SPI_SPEED_8MHZ);

RF24 radio(25,0, BCM2835_SPI_SPEED_1MHZ);

#define NRF24_CHANEL  100 

const uint8_t pipes[][6] = {{0xe1,0xe7,0xe7,0xe7,0xe7},{0xd2,0xe7,0xe7,0xe7,0xe7}};

struct type_packet_NRF24
{
  unsigned long stime;
  uint16_t  Vcc;
  int16_t  Temp;
  uint16_t  relH_H;
  uint16_t  relH_L;
  int16_t  absH;
  int16_t  TempH;
  uint16_t  Hum;
  uint16_t rain;
  uint16_t light;
  uint8_t error;
  uint8_t count;
  char note[6];
} 
packet;

void radio_setup()
{
  radio.begin();
  radio.csDelay=5;
  delay(2);
  radio.setDataRate(RF24_250KBPS);
  radio.setPALevel(RF24_PA_MAX);
  radio.setChannel(NRF24_CHANEL);
  radio.setCRCLength(RF24_CRC_16);
  radio.setAutoAck(true);
  //radio.enableDynamicPayloads();
  radio.enableAckPayload();
  radio.setRetries(15,15);

  //radio.printDetails();

  radio.openWritingPipe(pipes[1]);
  radio.openReadingPipe(1,pipes[0]);
  radio.startListening();
}

int main(int argc, char** argv)
{
  radio_setup();
  delay(20);

  radio.stopListening();
  delay(2);

  struct tm *ptr;
  time_t lt;
  lt = time(NULL);
  unsigned long stime = lt;
  ptr = localtime(&lt);
  char buffer [80];
  strftime (buffer,80,"%a %d %b %H:%M:%S %Z %Y",ptr);

  if(radio.failureDetected == 1)
  {
    delay(50);
    radio.failureDetected = 0;
    fprintf(stderr,"%s : Error. FAILURE_DETECTED...Restart\n", buffer);
    radio_setup();
    delay(20);

    radio.stopListening();
    delay(2);
  }

  int t=0;
  bool ok=0, rf_failure=0;
  while(t<10 && !ok)
  {
    ok = radio.write( &stime, sizeof(unsigned long) );
    radio.txStandBy();

    if(radio.failureDetected == 1)
    {
      radio.failureDetected = 0;
      if(rf_failure==1)
      {
        fprintf(stderr,"%s : Error. FAILURE_DETECTED...Exit\n", buffer);
        return 25;
      }
      else
      {
	delay(50);
        fprintf(stderr,"%s : Error. FAILURE_DETECTED...Restart\n", buffer);
        radio_setup();
	delay(20);

        radio.stopListening();
        delay(2);
      }
      rf_failure=1;
    }
    t++;
  }

  if (!ok)
  {
    fprintf(stderr,"%s : Error. Transmission_failed T=%d.\n",buffer,t);
    return 25;
  }
  else
  {
    radio.startListening();
    delay(2);

    unsigned long started_waiting_at = millis();
    unsigned long tof = millis()-started_waiting_at;

    bool timeout = false;
    while ( !radio.available() && !timeout )
    {
      if(radio.failureDetected == 1)
      {
        radio.failureDetected = 0;
        fprintf(stderr,"%s : Error. Receive_FAILURE_DETECTED...Restart\n", buffer);
        radio_setup();
      }

      tof = millis()-started_waiting_at;
      if (tof > 150)timeout = true;
    }

    if (timeout)
    {
      fprintf(stderr,"%s : Error. Response_timed_out T=%d time=%lu.\n",buffer,t,tof);
      return 25;
    }
    else
    {
      radio.read( &packet, sizeof(packet) );
      if(sizeof(packet)!=32 || packet.error!=0 || stime!=packet.stime || tof>300)
      {
	printf("%s : Error. Data_corrupted.\n",buffer);
	fprintf(stderr,"%d %lu %d %1.3f %2.2f %4.3f %d %2.2f %2.2f %d %d %d %s %d %lu\n",
        sizeof(packet),
        packet.stime,
        packet.error,
        packet.Vcc/1000.0F,
        packet.Temp/100.0F,
        packet.relH_H*10+packet.relH_L/1000.0F,
        packet.absH,
        packet.TempH/100.0F,
        packet.Hum/100.0F,
        packet.rain,
        packet.light,
        packet.count,
        packet.note,
        t,
        tof);
	return 25;
      }
      else
      {
        fprintf(stdout,"%d %lu %d %1.3f %2.2f %4.3f %d %2.2f %2.2f %d %d %d %s %d %lu\n",
        sizeof(packet),
        packet.stime,
        packet.error,
        packet.Vcc/1000.0F,
        packet.Temp/100.0F,
        packet.relH_H*10+packet.relH_L/1000.0F,
        packet.absH,
        packet.TempH/100.0F,
        packet.Hum/100.0F,
        packet.rain,
        packet.light,
        packet.count,
        packet.note,
        t,
        tof);
	return 0;
      }
    }
  }
  return 25;
}

