#include <SPI.h>
#include <Wire.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include "nRF24L01.h"
#include "RF24.h"
#include "printf.h"
#include "Adafruit_Sensor.h"
#include "Adafruit_BMP280.h"
#include <SI7021.h>

// Макросы для работы с портами  скорость и место
#define SetOutput(port,bit)       DDR ## port |= _BV(bit)
#define SetInput(port,bit)        DDR ## port &= ~_BV(bit)
#define SetBit(port,bit)          PORT ## port |= _BV(bit)
#define ClearBit(port,bit)        PORT ## port &= ~_BV(bit)
#define WritePort(port,bit,value) PORT ## port = (PORT ## port & ~_BV(bit)) | ((value & 1) << bit)
#define ReadPort(port,bit)        (PIN ## port >> bit) & 1
#define PullUp(port,bit)          { SetInput(port,bit); SetBit(port,bit); }
#define Release(port,bit)         { SetInput(port,bit); ClearBit(port,bit); }

//#define DEBUG

#define NRF24_CHANEL  100

byte addresses[][6] = {{0xe1, 0xe7, 0xe7, 0xe7, 0xe7}, {0xd2, 0xe7, 0xe7, 0xe7, 0xe7}};
unsigned long tt __attribute__((section(".noinit")));
struct type_packet_NRF24
{
  unsigned long stime;
  uint16_t  Vcc;
  int16_t  Temp;
  uint16_t  relH_H;
  uint16_t  relH_L;
  int16_t  absH;
  int16_t  TempH;
  uint16_t  Hum;
  uint16_t rain;
  uint16_t light;
  uint8_t error;
  uint8_t count;
  char note[6];
}
packet __attribute__((section(".noinit")));
RF24 radio(9, 10);

unsigned long times = 0;
unsigned long t = 0;
uint8_t sens = 0;

volatile uint8_t f_wdt = 0;

const long ConstADC = 1126400;

Adafruit_BMP280 bmp;

SI7021 htu;

void(* resetFunc) (void) = 0;

void setup()
{
  setup_watchdog(3);
  //wdt_disable();
  reset_data();

#ifdef  DEBUG
  Serial.begin(115200);
  printf_begin();
  Serial.println(F("Weather station debug"));
#endif
  if (setRadio())
  {
#ifdef  DEBUG
    radio.printDetails();
#endif
  }
  else
  {
    packet.error |= _BV(0);
#ifdef  DEBUG
    Serial.println(F("nrf24 fail"));
#endif
  }

  if (!bmp.begin(0x76))
  {
    packet.error |= _BV(1);
#ifdef  DEBUG
    Serial.println(F("Could not find a valid BMP280 sensor, check wiring!"));
#endif
  }
  else
  {
#ifdef  DEBUG
    Serial.println(F("BMP280 is OK"));
#endif
  }

  if (!htu.begin())
  {
    packet.error |= _BV(2);
#ifdef  DEBUG
    Serial.println(F("Could not find a valid HTU21 sensor, check wiring!"));
#endif
  }
  else
  {
    htu.setPrecision(0x01);
#ifdef  DEBUG
    Serial.println(F("HTU21 is OK"));
#endif
  }

  radio.stopListening();
  delayMicroseconds(400);
  radio.writeBlocking(&packet, sizeof(packet), 200);
  radio.txStandBy();
  powerdownSPI();

#ifdef DEBUG
  Serial.print(F("Pressure = "));
  Serial.print(bmp.readPressure());
  Serial.println(" Pa");

  Serial.print(F("Temperature = "));
  Serial.print(bmp.readTemperature());
  Serial.println(" C");

  Serial.print(F("Approx altitude = "));
  Serial.print(bmp.readAltitude(1013.25));
  Serial.println(" m");

  si7021_env data = htu.getHumidityAndTemperature();
  Serial.print(F("Humidity = "));
  Serial.print(data.humidityBasisPoints / 100.0F);
  Serial.println(" %");

  Serial.print(F("Temperature = "));
  Serial.print((float)(data.celsiusHundredths / 100.0F));
  Serial.println(" C");

  Serial.print(F("Vcc = "));
  Serial.print(readVcc() / 1000.0F);
  Serial.println(" V");

  Serial.print(F("Rain = "));
  Serial.print(readRainSensor());
  Serial.println(" p");

  Serial.print(F("Sun = "));
  Serial.print(readSunSensor());
  Serial.println(" p");
#endif

  //setup_watchdog(3);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN); // здесь устанавливается режим сна
  sleep_enable();
  sleep_mode();                        // система засыпает
  sleep_disable();                    // система продолжает работу после переполнения watchdog
}

void loop()
{
  if (f_wdt == 1)
  {
    times = millis();

    powerupSPI();
    radio.startListening();

    unsigned long got_time;
    unsigned long timeEnd = micros() + 4000;

    while (micros() < timeEnd)
    {
      if (radio.available())
      {
        while (radio.available())
        {
          radio.read(&got_time, 4);
        }
        packet.stime = got_time;

        packet.note[4] = 5 + '0';
        si7021_env data = htu.getHumidityAndTemperature();//6-7ms
        packet.TempH = data.celsiusHundredths;
        packet.Hum = data.humidityBasisPoints;

        sendRadio();
        break;
      }
      else if (timeEnd - micros() > 3500)
      {
        readSensors();
      }
    }
    powerdownSPI();

    f_wdt = 0;
    system_sleep();
  }
  else
  {
    system_sleep();
  }
}

bool setRadio(void)
{
  if (radio.begin() == false) radio.begin();
  //radio.begin();
  radio.setDataRate(RF24_250KBPS);
  radio.setPALevel(RF24_PA_MAX);
  radio.setChannel(NRF24_CHANEL);
  radio.setCRCLength(RF24_CRC_16);
  radio.setAutoAck(true);
  //radio.setAutoAck(false);
  //radio.enableDynamicPayloads();
  radio.enableAckPayload();
  radio.setRetries(4, 15);
  radio.openWritingPipe(addresses[0]);
  radio.openReadingPipe(1, addresses[1]);
  radio.startListening();
  //radio.powerDown();
  return true;
}

void powerdownSPI()
{
  radio.powerDown();
  SPI.end();
  PullUp(B, 5);
  PullUp(B, 3);
}

void powerupSPI()
{
  SetOutput(B, 5);
  SetOutput(B, 3);
  SPI.begin();
  radio.powerUp();
}

bool sendRadio(void)
{
  bool send_packet_ok = false;
  radio.stopListening();
  delayMicroseconds(400);

  t = millis() - times;
  packet.note[4] = '7';
  packet.note[0] = t / 100 + '0';
  packet.note[1] = (t - (t / 100)) / 10 + '0';
  packet.note[2] = (t - (t / 100)) % 10 + '0';
  t = 0;

  radio.writeBlocking(&packet, sizeof(packet), 200);
  send_packet_ok = radio.txStandBy();
  if (radio.isAckPayloadAvailable())radio.read(&t, sizeof(t));

  delay(1);

  packet.error = 0x00;
  packet.count++;
}

void reset_data(void)
{
  //packet.error=0;

  packet.Vcc = 9999;
  packet.Temp = 9999;
  packet.relH_H = 9999;
  packet.relH_L = 9999;
  packet.absH = 9999;
  packet.TempH = 9999;
  packet.Hum = 9999;
  packet.rain = 9999;

  packet.note[0] = '_';
  packet.note[1] = '_';
  packet.note[2] = '_';
  packet.note[3] = 's';
  packet.note[4] = '0';
  packet.note[5] = 0;

  packet.stime = 0;
  packet.count = 0;
}

void system_sleep()
{
  packet.note[4] = '8';
  ADCSRA &= ~(_BV(ADEN));
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);
  sleep_enable();
  sleep_mode();
  sleep_disable();
  ADCSRA |= (_BV(ADEN));
}

//**********************************************
// 0=16ms, 1=32ms,2=64ms,3=128ms,4=250ms,5=500ms
// 6=1 sec,7=2 sec, 8=4 sec, 9= 8sec
void setup_watchdog(int ii)
{
  byte bb;
  if (ii > 9 ) ii = 9;
  bb = ii & 7;
  if (ii > 7) bb |= (1 << 5);
  bb |= _BV(WDCE);
  MCUSR &= ~(_BV(WDRF));
  WDTCSR |= _BV(WDCE) | _BV(WDE);
  WDTCSR = bb;
  WDTCSR |= _BV(WDIE);
}

ISR(WDT_vect)
{
  if (f_wdt == 1)
  {
    packet.error |= _BV(3);
    t = millis() - times;
    packet.note[4] = (char)('9');
    packet.note[0] = (char)(t / 100 + '0');
    packet.note[1] = (char)((t - (t / 100)) / 10 + '0');
    packet.note[2] = (char)((t - (t / 100)) % 10 + '0');

    resetFunc();
  }
  f_wdt = 1;
}

long readVcc()
{
  long result;
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delayMicroseconds(400);
  ADCSRA |= _BV(ADSC); // Convert
  unsigned long timeEnd = micros() + 1000;
  while (bit_is_set(ADCSRA, ADSC) && micros() < timeEnd);
  result = ADCL;
  result |= ADCH << 8;
  result = ConstADC / result; // Back-calculate AVcc in mV
  return result;
}

long readRainSensor()
{
  long result;
  ADMUX = _BV(REFS0);
  delayMicroseconds(400);
  ADCSRA |= _BV(ADSC); // Convert
  unsigned long timeEnd = micros() + 1000;
  while (bit_is_set(ADCSRA, ADSC) && micros() < timeEnd);
  result = ADCL;
  result |= ADCH << 8;
  return result;
}

long readSunSensor()
{
  long result;
  ADMUX = _BV(REFS0) | _BV(MUX0);
  delayMicroseconds(400);
  ADCSRA |= _BV(ADSC); // Convert
  unsigned long timeEnd = micros() + 1000;
  while (bit_is_set(ADCSRA, ADSC) && micros() < timeEnd);
  result = ADCL;
  result |= ADCH << 8;
  //result = ConstADC / result;
  return result;
}

void readSensors()
{
  float pressure;
  packet.note[4] = (char)(sens / 2 + '0');

  switch (sens)
  {
    case 0:
      delay(1);
      break;

    case 1:
      packet.light = (uint16_t)readSunSensor(); //1-2ms
      break;

    case 2:
      packet.Temp = (int16_t) (bmp.readTemperature() * 100); //1-2ms
      break;

    case 3:
      packet.Vcc = (uint16_t)readVcc(); //1-2ms
      break;

    case 4:
      packet.absH = (int16_t) bmp.readAltitude(1013.25); //2-3ms
      break;

    case 5:
      packet.rain = (uint16_t)readRainSensor(); //1-2ms
      break;

    case 6:
      pressure = bmp.readPressure(); //2-4ms
      packet.relH_H = (uint16_t) (pressure / 1000.0F);
      packet.relH_L = (uint16_t) (((long)(pressure * 10)) % 10000);
      break;

    default:
      sens = 0;
      break;
  }
  sens++;
}









